# SWFFile - an Image-like object for SWF files
# Copyright 2004 Nick Evan-Wong

__version__ = 0.31

from OFS import PropertyManager, SimpleItem
from OFS.PropertyManager import PropertyManager
from OFS.Image import Image
from OFS.SimpleItem import Item
from AccessControl import ClassSecurityInfo, getSecurityManager, Role
from AccessControl.Role import RoleManager
from Products.PageTemplates.PageTemplateFile import PageTemplateFile
from Globals import InitializeClass
from Globals import Persistent
from OFS.Cache import Cacheable


# Management interface methods and declarations
# ---------------------------------------------
addSWFFileForm=PageTemplateFile('zpt/addSWF', globals())

# ZMS constructor method
def manage_addSWFFile(
	self,
	id,
	title='',	
	file=None,
	REQUEST=None
):
	"""ZMS constructor"""
			
	id = str(id)
	title = str(title)			
		
	Obj = SWFFile(id,title,file=file)
	
	self = self.this()
	self._setObject(id, Obj)
				
	if REQUEST is not None:				
		return REQUEST.RESPONSE.redirect('%s/manage_main' % self._getOb(id).absolute_url())



# Implement the SWFFile Class
# ---------------------------
class SWFFile(Image, Cacheable):
	"""Implementation of the SWFFile class"""
	
	meta_type='SWF File'	

	# define properties
	getQuality = ['autolow','low','high','autohigh','best']
	getScale = ['showall','noborder','exactfit']
	getAlign = ['l','r','t','b']
	getWmode = ['window','opaque','transparent']

	_properties = (
			{'id':'title',			'type':'string',	'mode':'w'},
			{'id':'movieid', 		'type':'string',	'mode':'w'},
	        {'id':'bgcolor', 		'type':'string', 	'mode':'w'},
	        {'id':'base', 			'type':'string', 	'mode':'w'},			
			{'id':'width', 			'type':'int', 		'mode':'w'},
	        {'id':'height', 		'type':'int', 		'mode':'w'},	        
	        {'id':'align', 			'type':'selection', 'mode':'w', 'select_variable' : 'getAlign'},
	        {'id':'quality', 		'type':'selection', 'mode':'w', 'select_variable' : 'getQuality'},
	        {'id':'scale', 			'type':'selection', 'mode':'w',	'select_variable' : 'getScale'},	        
	        {'id':'wmode', 			'type':'selection', 'mode':'w',	'select_variable' : 'getWmode'},
	        {'id':'flashvars', 		'type':'lines', 	'mode':'w'},
			{'id':'play',			'type':'boolean', 	'mode':'w'},
	        {'id':'loop', 			'type':'boolean', 	'mode':'w'},
	        {'id':'menu', 			'type':'boolean', 	'mode':'w'},	        	        
	        {'id':'devicefont',    	'type':'boolean', 	'mode':'w'},	        
			{'id':'swliveconnect', 	'type':'boolean', 	'mode':'w'},			
	        {'id':'content_type', 	'type':'string', 	'mode':''},
	        {'id':'flashversion', 	'type':'string', 	'mode':''},
	        {'id':'compressed', 	'type':'boolean', 	'mode':''},	        
	        {'id':'frame_rate', 	'type':'float', 	'mode':''},
	        {'id':'frame_count', 	'type':'int', 		'mode':''}
	)

	
	
	# define management templates		
	manage_view=PageTemplateFile('zpt/viewSWF', globals())

	manage_options=(
        (        		
        	{'label':'Edit', 'action':'manage_main' },
        	{'label':'View', 'action':'manage_view' },
        )
		+ PropertyManager.manage_options
        + RoleManager.manage_options
        + Item.manage_options
		+ Cacheable.manage_options 
        )
				
	__ac_permissions__=(
		('View management screens',('manage', 'manage_main')),
		('Upload SWFFile data',	('manage_upload','PUT')),
 		('Change SWFFile properties', ('manage_addProperty',
                               'manage_editProperties',
                               'manage_delProperties',
                               'manage_changeProperties',)),
		('View',('index_html', 'manage_view', 'view_image_or_file', 'get_size', '')),
    	('FTP access', ('manage_FTPstat','manage_FTPget','manage_FTPlist')),
    	('Delete objects',('DELETE',)),
    )
	
	security=ClassSecurityInfo()
			

	def __init__(self, id,
		title='',
		movieid='',	
		play=1,
		loop=0,
		menu=1,
		swliveconnect=0,
		quality='high',
		scale='showall',
		wmode='window',
		align='',
		bgcolor='',
		devicefont=0,
		base='',
		flashvars='',
		file=None,
		REQUEST=None):
		"initialise"
				
		# workaround OFS.SimpleItem.Item_w__name__ bug in Zope 2.7.2
		self.__name__ = ''
		self._setId(id)
		
		self.title = title
			
		if not movieid:
			movieid = id
		self.movieid = movieid
		self.play = play
		self.loop = loop
		self.menu = menu
		self.swliveconnect = swliveconnect
		self.quality = quality
		self.scale = scale
		self.wmode = wmode
		self.align = align
		self.bgcolor = bgcolor
		self.devicefont = devicefont

		self.base = base
		self.flashvars = flashvars
		self.flashversion = ''
		self.compressed = ''
		self.content_type = 'application/x-shockwave-flash'	
		
		self.frame_count = 0
		self.frame_rate = 0
	
		# upload the SWF file if present
		if file is not None:
			self.manage_upload(file=file)
					

	def __str__(self):
		"Return the string representation"
		return self.tag()
		
	
  	def __setstate__(self,state):
		
		Persistent.__setstate__(self,state)
		
		if not hasattr(self,'play'):
                	try: play = self.autoplay
                	except: play = 1
                	self.play = play
                	
		# new properties added in 0.3
		if not hasattr(self,'frame_count'):
			self.frame_count = 0    				    			
		if not hasattr(self,'frame_rate'):
			self.frame_rate = 0			
		if not hasattr(self,'devicefont'):
			self.devicefont = 0
		if not hasattr(self,'base'):
			self.base = ''
		if not hasattr(self,'flashvars'):
			self.flashvars = ''
		if not hasattr(self,'flashversion'):
			self.flashversion = ''
    				
    						
 	def tag(self,
		height=None, 
		width=None,
		scale=0,
		xscale=0,
		yscale=0, **args):
		"HTML tag"

		# Scale SWFFile - from OFS.Image
		if height is None: height=str(self.height)
		if width is None:  width=str(self.width)

        # Auto-scaling support
		xdelta = xscale or scale
		ydelta = yscale or scale

		if xdelta and width:
			width = str(int(round(int(width) * xdelta)))
		if ydelta and height:
			height = str(int(round(int(height) * ydelta)))

		if self.loop: loop = 'true'
		else: loop = 'false'
		if self.menu: menu = 'true'
		else: menu = 'false'
		if self.swliveconnect: swliveconnect = 'true'
		else: swliveconnect = 'false'
		if self.play: play = 'true'
		else: play = 'false'
		if self.devicefont: devicefont = 'true'
		else: devicefont = 'false'

		query_string = ''
		# generate flashvars query string
		if self.flashvars>0:
			from string import split
			from urllib import quote
			
			for pair in self.flashvars:
				try:
					(key,val) = split(pair,'=')
					query_string += key + '=' + val + '&'
				except:
					query_string = ''

			# remove trailing &
			query_string = query_string[:-1]
			# encode string, except & and = chars
			query_string = quote(query_string,'&,=')
			
	        tag =  '\n<object classid=\"clsid:D27CDB6E-AE6D-11cf-96B8-444553540000\" codebase=\"http://download.macromedia.com/pub/shockwave/cabs/flash/swflash.cab#version=6,0,0,0\" id=\"%s\" width=\"%s\" height=\"%s\" >\n' % (self.movieid,width, height)
	        tag += '\t<param name=\"movie\" value=\"'+self.absolute_url()+'\" />\n'
	        tag += '\t<param name=\"base\" value=\"'+self.base+'\" />\n'
	        tag += '\t<param name=\"quality\" value=\"'+self.quality+'\" />\n'
	        tag += '\t<param name=\"scale\" value=\"'+self.scale+'\" />\n'	        
	        tag += '\t<param name=\"play\" value=\"'+play+'\" />\n'
	        tag += '\t<param name=\"loop\" value=\"'+loop+'\" />\n'
	        tag += '\t<param name=\"menu\" value=\"'+menu+'\" />\n'
	        tag += '\t<param name=\"swliveconnect\" value=\"'+swliveconnect+'\" />\n'
	        tag += '\t<param name=\"devicefont\" value=\"'+devicefont+'\" />\n'
	        tag += '\t<param name=\"bgcolor\" value=\"'+self.bgcolor+'\" />\n'
	        tag += '\t<param name=\"align\" value=\"'+self.align+'\" />\n'
	        tag += '\t<param name=\"wmode\" value=\"'+self.wmode+'\" />\n'
	        tag += '\t<param name=\"flashvars\" value=\"'+query_string+'\" />\n'
	        tag += '\t<embed src=\"%s\" quality=\"%s\" bgcolor=\"%s\" width=\"%s\" height=\"%s\" name=\"%s\" align=\"%s\" type=\"application/x-shockwave-flash\" PLUGINSPAGE=\"http://www.macromedia.com/go/getflashplayer\"></embed>\n' % (self.absolute_url(), self.quality, self.bgcolor, width, height, self.movieid, self.align)
	        tag += '</object>\n'
			
        	return tag


	def manage_upload(self,file=None,title=None,REQUEST=None):
		"Upload file data and set title"

		if self.wl_isLocked():
			raise ResourceLockedError, "File is locked via WebDAV"
		
		if title:
			self.title = title
			message = 'Set title'														
			
		if file:					
			data, size = self._read_data(file)				
			file.seek(0)
			data = file.read()
			filesize = len(data)
			analysed = self.analyseContent(data,filesize)						
			
			# set analysed property values		
			self.height = analysed['height']
			self.width = analysed['width']		
			self.flashversion = analysed['flashversion']
			self.compressed = analysed['compressed']
			self.frame_count = analysed['frame_count']
			self.frame_rate = analysed['frame_rate']
			# update data
			self.update_data(data, self.content_type, size)
				
			if REQUEST is not None:
				try:    url=self.DestinationURL()
				except: url=REQUEST['URL1']
				REQUEST.RESPONSE.redirect('%s/manage_main' % url)
			

	# get unsigned bits
	def _getUBits(self, bitList, bitPos, nr):
		valD = 0
		for runD in xrange(bitPos, bitPos+nr):
			valD = valD*2 + bitList[runD]				
		return valD

	# get signed bits
	def _getSBits(self, bitList, bitPos, nr):
		valD = 0
		for runD in xrange(bitPos+1, bitPos+nr):
			valD = valD*2 + bitList[runD]    			
		if bitList[bitPos]: 
			valD = -valD
		return valD
		
		
		
	# analyse SWF file content
	def analyseContent(self, data, filesize):

		if not data: return None 

		if len(data)<=(8+17):
			raise "Content error","data too small!"
		# Uncompressed flash file
		if str(data[:3]) == 'FWS':
			compressed = 0
		# ZLIB compressed flash file
		elif ((str(data[:3]) == 'CWS') or (str(data[:3]) == 'FWC')):
			compressed = 1
		else:
			raise "Content Error","This does not appear to be an SWF file!"
    	
		fversion = str(ord(data[3]))
		size = ord(data[4]) + ord(data[5])*256 + ord(data[6])*256*256 + ord(data[7])*256*256*256    		
		data = data[8:1024]
		
		if compressed:
			from zlib import decompressobj
			do = decompressobj()
			data = do.decompress(data)
			do.flush()
    
		bitList = []
    	
		for runD in xrange(0, 17):
			byteD = ord(data[runD])
			bitList.append( (byteD&128)/128 )
			bitList.append( (byteD&64)/64 )
			bitList.append( (byteD&32)/32 )
			bitList.append( (byteD&16)/16 )
			bitList.append( (byteD&8)/8 )
			bitList.append( (byteD&4)/4 )
			bitList.append( (byteD&2)/2 )
			bitList.append( (byteD&1)/1 )        
    
		bitPos = 0
		fieldSize = self._getUBits(bitList, bitPos, 5); bitPos=bitPos+ 5
		x1 = self._getSBits(bitList, bitPos, fieldSize); bitPos=bitPos+ fieldSize
		x2 = self._getSBits(bitList, bitPos, fieldSize); bitPos=bitPos+ fieldSize
		y1 = self._getSBits(bitList, bitPos, fieldSize); bitPos=bitPos+ fieldSize
		y2 = self._getSBits(bitList, bitPos, fieldSize); bitPos=bitPos+ fieldSize

		p=(bitPos/8+1)
		frameRateMinor=ord(data[p]); p+=1
		frameRateMajor=ord(data[p]); p+=1
		framesTotal=ord(data[p])+ord(data[p+1])*254

		return {
    		'filesize'		:	filesize,
    		'flashversion'	:	fversion,
    		'compressed'	:	compressed,
			'size'			:	size,
			'width'			: 	int(x2)/20,
        	'height'		:	int(y2)/20,
        	'frame_rate'	:	'%s.%s' % (frameRateMajor, frameRateMinor),
        	'frame_count'	:	framesTotal
    	}


# Initialize product class
# ------------------------
InitializeClass(SWFFile)
